
#include <iostream>
#include <fstream>
#include <math.h>
#include <time.h>
#include "fcmanager.h"

double g_lambda = 3;
double g_lambda_r = 3;
double g_reliability = 1;
bool l_init = false;
double l_fact[FACTORIAL_MAX+1];

int getcharcount(char c, string& s)
{
	int i, len;
	int count = 0;
	for (i=0, len = s.length(); i<len; i++)
		if (s.at(i)==c)
			count++;
	return count;
}

double modulo(double val)
{
	if (val<0) return -val;
	return val;
}

double log_factorial(int i)
{
	if (!l_init)
	{
		for (int i=0; i<=FACTORIAL_MAX; i++)
			l_fact[i] = -1;
		l_init = true;	
	}
	if (i<=1)
		return 0;
	if (l_fact[i]>=0)
		return l_fact[i];
	l_fact[i] = log_factorial(i-1)+log(i);
	return l_fact[i];
}

double factorial(int i)
{
	return exp(log_factorial(i));
}

double combinatorial(int N, int k)
{
	return exp(log_factorial(N)-log_factorial(N-k)-log_factorial(k));
}

double log_combinatorial(int N, int k)
{
	return log_factorial(N)-log_factorial(N-k)-log_factorial(k);
}

char *trimleft(char* string)
{
	while(*string==' ' || *string=='\t')
		string++;
	return string;
}


char* tokenize(char *string, char token, int &pos, char *tokenbuf)
{
	int i=0;
	while(string[pos]==token)
		pos++;
	while(string[pos]!=token && string[pos]!='\0')
		tokenbuf[i++] = string[pos++];
	tokenbuf[i]='\0';
	return tokenbuf;
}

//##################### FC ########################

FC::FC(short id, string &name, FC* parent)
{
	_id = id;
	_name = name;
	_parent = parent;
	_occ = 0;
	_roc = 1;
}

FC::~FC()
{
	_parent = NULL;
	_orfs.RemoveAll();
	_children.RemoveAll();
}

string FC::getFC()
{
	string fc;
	char buffer[5];
	sprintf(buffer, "%d", _id);
	fc = buffer;

	FC *current = _parent;
	while(current)
	{
		sprintf(buffer, "%d", current->_id);
		fc = string(buffer) + "." + fc;
		current = current->_parent;
	}
	return fc;
}

void FC::removeORF(ORF* annot)
{
	_orfs.RemoveKey(annot->_name);
}

bool FC::haveORF(ORF* annot)
{
	return _orfs.Lookup(annot->_name)!=NULL;
}

//##################### ORF ########################

ORF::ORF(string &name, FC *fc)
{
//		_assignedFC = NULL;
	_name = name;
	if (fc) _fc.push_back(fc);
	_score = 0;
}

ORF::~ORF()
{
	FCList::iterator it;
	for(it = _fc.begin(); it != _fc.end(); it++)
	{
		if ((*it)->_orfs.GetCount()>0)
			(*it)->removeORF(this);
	}
	_fc.clear();

	POSITION pos;
	for(pos = _neighbours.GetHeadPosition(); pos;)
	{
		NeighbourRelation *rel;
		ORF *neighbour;
		_neighbours.GetNextAssoc(pos, neighbour, rel);
		delete rel;
		neighbour->removeNeighbour(this);
	}
	_neighbours.RemoveAll();
}

void ORF::removeFC(FC* fc)
{
	FCList::iterator it = _fc.begin();
	while(it != _fc.end())
	{
		if (*it == fc)
		{
			_fc.erase(it);
			return;
		}
		it++;
	}		
}

int ORF::addNeighbour(ORF* orf, string &type, double weight)
{
	int count;
	NeighbourRelation *rel;
	if (_neighbours.Lookup(orf, rel))
	{
		if (!rel->_types.Lookup(type, count))
			count = 0;
		count++;
		rel->_types.SetAt(type, count);
		return 1;
	}
	_neighbours.SetAt(orf, new NeighbourRelation(type, weight));
	return 0;
}

void ORF::removeNeighbour(ORF* orf)
{
	NeighbourRelation *rel;
	if (_neighbours.Lookup(orf, rel))
		delete rel;	
	_neighbours.RemoveKey(orf);
}

bool ORF::hasFC(FC* fc)
{
	FCList::iterator it;
	for (it = _fc.begin(); it != _fc.end(); it++)
		if (*it == fc)
			return true;
	return false;
}

bool ORF::hasOFC(FC* fc)
{
	FCList::iterator it;
	for (it = _ofc.begin(); it != _ofc.end(); it++)
		if (*it == fc)
			return true;
	return false;
}

bool ORF::hasBackupFC(FC* fc)
{
	FCList::iterator it;
	for (it = _backupfc.begin(); it != _backupfc.end(); it++)
		if (*it == fc)
			return true;
	return false;
}

int ORF::getSharedFCcount(ORF* annot)
{
	int count = 0;
	FCList::iterator it = _fc.begin();
	while(it != _fc.end())
	{
		if (annot->hasFC(*it))
			count++;
		it++;
	}
	return count;
}

double ORF::getSharedL2FCWeight(ORF* orf)
{
	int count, shared = 0, total = 0;
	CRBMap <FC*, int> x;
	FCList::iterator it;
	for(it = _fc.begin(); it != _fc.end(); it++)
		x.SetAt(*it, 1);

	POSITION pos;
	for (pos = orf->_neighbours.GetHeadPosition(); pos;)
	{
		ORF *neighbour = orf->_neighbours.GetNextKey(pos);
		if (neighbour == this)
			continue;
		for(it = neighbour->_fc.begin(); it != neighbour->_fc.end(); it++)
		{
			if (!x.Lookup(*it, count))
				count = 0;
			count++;
			x.SetAt(*it, count);		
		}
	}
	
	pos = x.GetHeadPosition();
	while(pos)
	{
		if (x.GetNextValue(pos)>1)
			shared++;
	}
	return (double)shared/x.GetCount();
}

double ORF::getSharedFCWeight(ORF* annot)
{
	int count, shared = 0, total = 0;
	CRBMap <FC*, int> x;
	FCList::iterator it;
	for(it = _fc.begin(); it != _fc.end(); it++)
		x.SetAt(*it, 1);
	for(it = annot->_fc.begin(); it != annot->_fc.end(); it++)
		if (x.Lookup(*it, count))
			shared++;	
	return (double)shared/(_fc.size() + annot->_fc.size() - shared);
}

double ORF::getNeighbourSharedFCWeight(ORF* orf)
{
	int count = 0;
	double weight = 0;
	POSITION pos;
	for(pos = _neighbours.GetHeadPosition(); pos;)
	{
		ORF *neighbour = _neighbours.GetNextKey(pos);
		if (neighbour == orf) continue;
		weight += getSharedFCWeight(neighbour);
		count++;
	}
	if (count == 0)
		return 1;
	return weight /= count;
}

double ORF::getSharedNeighbourWeight(ORF* orf)
{
	int count, shared = 0, total = 0;
	CRBMap <ORF*, int> x;
	POSITION pos;
	for(pos = _neighbours.GetHeadPosition(); pos;)
	{
		ORF *neighbour = _neighbours.GetNextKey(pos);
		if (neighbour == orf)
		{
			shared++;
			total++;
		}else
			x.SetAt(neighbour, 1);
	}
	for(pos = orf->_neighbours.GetHeadPosition(); pos;)
	{
		ORF *neighbour = orf->_neighbours.GetNextKey(pos);
		if (neighbour == this)
		{
			shared++;
			total++;
		}else
		if (x.Lookup(neighbour, count))
			shared++;
	}
	total += _neighbours.GetCount() + orf->_neighbours.GetCount() - shared;
	return (double)shared/total;
}

double ORF::computeCWeight(ORF *orf)
{
	FCList::iterator ft;
	int count, total = 1;
	CRBMap <FC *, int> map;
	POSITION pos;
	for (pos = _neighbours.GetHeadPosition(); pos;)
	{
		ORF *neighbour = _neighbours.GetNextKey(pos);
		if (neighbour->_fc.size()==0) continue;
		total++;
		for (ft = neighbour->_fc.begin(); ft != neighbour->_fc.end(); ft++)
		{
			if (!map.Lookup(*ft, count))
				count = 0;
			count++;
			map.SetAt(*ft, count);
		}
	}
	double weight1 = 0;
	for (ft = orf->_fc.begin(); ft != orf->_fc.end(); ft++)
	{
		if (!map.Lookup(*ft, count))
			count = 0;
		weight1 = max(weight1, (double)count/total);
	}
	if (orf->_fc.size()==0)
		weight1 = 0.5;
	total = 0;
	map.RemoveAll();
	for (pos = orf->_neighbours.GetHeadPosition(); pos;)
	{
		ORF *neighbour = orf->_neighbours.GetNextKey(pos);
		if (neighbour->_fc.size()==0) continue;
		total++;
		for (ft = neighbour->_fc.begin(); ft != neighbour->_fc.end(); ft++)
		{
			if (!map.Lookup(*ft, count))
				count = 0;
			count++;
			map.SetAt(*ft, count);
		}
	}
	double weight2 = 0;
	for (ft = _fc.begin(); ft != _fc.end(); ft++)
	{
		if (!map.Lookup(*ft, count))
			count = 0;
		weight2 = max(weight2, (double)count/total);
	}
	if (_fc.size()==0)
		weight2 = 0.5;
	return weight1*weight2;
}

int ORF::getFCCount()
{
	int count = 0;
	FCList::iterator it;
	for (it = _fc.begin(); it != _fc.end(); it++)
	{
		if ((*it)->_orfs.GetCount()>0)
			count++;
	}
	return count;
}

double ORF::computeCD_sim(ORF *orf, bool reliability)
{
	CRBMap <ORF*, double> xNeighbours;
	double sim = 0;
	double base1 = 0;
	double base2 = 0;
	double weight;
	POSITION pos;
	for (pos = _neighbours.GetHeadPosition(); pos;)
	{
		NeighbourRelation *rel;
		ORF *neighbour;
		_neighbours.GetNextAssoc(pos, neighbour, rel);
		if (reliability)
		{
			xNeighbours.SetAt(neighbour, rel->_weight2);
			base1 += rel->_weight2;
		}else
		{
			xNeighbours.SetAt(neighbour, 1);
			base1 ++;
		}
	}

	if (xNeighbours.Lookup(orf, weight))
	{
		base1 -= weight;
		base1 += (1 - weight); 
		base2 += (1 - weight); 
		sim += weight*2;
	}else{
		base1 ++;
		base2 ++;
	}

	for (pos = orf->_neighbours.GetHeadPosition(); pos;)
	{
		NeighbourRelation *rel;
		ORF *neighbour;
		orf->_neighbours.GetNextAssoc(pos, neighbour, rel);
		
		if (neighbour == this) continue;
		if (xNeighbours.Lookup(neighbour, weight))
		{
			base1 -= weight;
			if (reliability)
			{
				base1 += (1 - rel->_weight2)*weight;
				base2 += (1 - weight)*rel->_weight2;
				sim += weight*rel->_weight2;
			}else
				sim ++;
		}else
		{
			if (reliability)
				base2 += rel->_weight2;
			else
				base2 ++;
		}
	}
	return ((sim*2)/(sim*2+base1))*((sim*2)/(sim*2+base2));
}

double ORF::computeFS_weight(ORF *orf, int N, bool reliability)
{
	CRBMap <ORF*, double> xNeighbours;
	double sim = 0;
	double base1 = 0;
	double base2 = 0;
	double weight;
	POSITION pos;
	for (pos = _neighbours.GetHeadPosition(); pos;)
	{
		NeighbourRelation *rel;
		ORF *neighbour;
		_neighbours.GetNextAssoc(pos, neighbour, rel);
		
		if (reliability)
		{
			xNeighbours.SetAt(neighbour, rel->_weight2);
			base1 += rel->_weight2;
		}else
		{
			xNeighbours.SetAt(neighbour, 1);
			base1 ++;
		}
	}

	if (xNeighbours.Lookup(orf, weight))
	{
		base1 -= weight;
		base1 += (1 - weight); 
		base2 += (1 - weight); 
		sim += weight*2;
	}else{
		base1 ++;
		base2 ++;
	}

	for (pos = orf->_neighbours.GetHeadPosition(); pos;)
	{
		NeighbourRelation *rel;
		ORF *neighbour;
		orf->_neighbours.GetNextAssoc(pos, neighbour, rel);
		if (neighbour == this) continue;
		if (xNeighbours.Lookup(neighbour, weight))
		{
			base1 -= weight;
			if (reliability)
			{
				base1 += (1 - rel->_weight2)*weight;
				base2 += (1 - weight)*rel->_weight2;
				sim += weight*rel->_weight2;
			}else
				sim ++;
		}else
		{
			if (reliability)
				base2 += rel->_weight2;
			else
				base2 ++;
		}
	}

	base1 = max(reliability?g_lambda_r:g_lambda, sim+base1);
	base1 -= sim;
	base2 = max(reliability?g_lambda_r:g_lambda, sim+base2);
	base2 -= sim;

//	return ((sim*sim)/(sim+base1))*((sim)/(sim+base2));
	return ((sim*2)/(sim*2+base1))*((sim*2)/(sim*2+base2));
}

double ORF::computeGeometric(ORF *orf, bool reliability)
{
	double Nv = _neighbours.GetCount();
	double Nw = orf->_neighbours.GetCount();
	double sim = 0;
//	if (_neighbours.Lookup(orf))
//		sim++;

	POSITION pos;
	for (pos = _neighbours.GetHeadPosition(); pos;)
	{
		if (orf->_neighbours.Lookup(_neighbours.GetNextKey(pos)))
			sim ++;
	}
	return (sim*sim)/(Nv*Nw);
}

double ORF::computeHyperGeometricScore(ORF *orf, int N)
{
	double Nv = _neighbours.GetCount();
	double Nw = orf->_neighbours.GetCount();
	int sim = 0;

	if (_neighbours.Lookup(orf))
	{
		Nv--;
		Nw--;
	}

	POSITION pos;
	for (pos = _neighbours.GetHeadPosition(); pos;)
	{
		if (orf->_neighbours.Lookup(_neighbours.GetNextKey(pos)))
			sim ++;
	}

	int k = min(Nv, Nw);
	double prob = 0;
	for (int i=sim; i<=k ;i++)
	{
//		prob += (combinatorial(N, i)*combinatorial(N - i, Nv - i)*combinatorial(N - Nv, Nw - i))/(combinatorial(N, Nv)*combinatorial(N, Nw));
		prob += exp(log_combinatorial(Nv, i)+log_combinatorial(N - Nv, Nw - i)-log_combinatorial(N, Nw));
	}
	prob = min(max(0, prob), 1);


	return -log(prob);
}

int ORF::getAnnotatedL1NeighbourCount()
{
	int annotated = 0;
	POSITION pos;
	for (pos = _neighbours.GetHeadPosition(); pos;)
		if (_neighbours.GetNextKey(pos)->_fc.size()>0)
			annotated++;
	return annotated;
}

int ORF::getAnnotatedOL1NeighbourCount()
{
	int annotated = 0;
	POSITION pos;
	for (pos = _neighbours.GetHeadPosition(); pos;)
		if (_neighbours.GetNextKey(pos)->_ofc.size()>0)
			annotated++;
	return annotated;
}

int ORF::getAnnotatedL2NeighbourCount()
{
	short count;
	CRBMap <ORF*, short> map;
	POSITION pos, pos1;
	for (pos = _neighbours.GetHeadPosition(); pos;)
	{
		ORF *neighbour = _neighbours.GetNextKey(pos);
		for (pos1 = neighbour->_neighbours.GetHeadPosition(); pos1;)
		{
			ORF *neighbour1 = neighbour->_neighbours.GetNextKey(pos1);
			if (neighbour1->_fc.size()>0)
			{
				if (!map.Lookup(neighbour1, count))
					map.SetAt(neighbour1, 1);
			}
		}
	}
	return map.GetCount();
}

double ORF::computeCD_dist_OLD(ORF *orf)
{
	CRBMap <ORF*, double> xNeighbours, yNeighbours, xyNeighbours;
	double weight, base = 0;
	xNeighbours.SetAt(this, 1);

	POSITION pos;
	for (pos = _neighbours.GetHeadPosition(); pos;)
	{
		NeighbourRelation *rel;
		ORF *neighbour;
		_neighbours.GetNextAssoc(pos, neighbour, rel);
		xNeighbours.SetAt(neighbour, rel->_weight2);
	}
	
	if (xNeighbours.Lookup(orf, weight))
	{
		xyNeighbours.SetAt(orf, weight);
		base += 1 - weight;
		xNeighbours.RemoveKey(orf);
	}else
		yNeighbours.SetAt(orf, 1);

	for (pos = orf->_neighbours.GetHeadPosition(); pos;)
	{
		NeighbourRelation *rel;
		ORF *neighbour;
		orf->_neighbours.GetNextAssoc(pos, neighbour, rel);

		if (xNeighbours.Lookup(neighbour, weight))
		{
			xyNeighbours.SetAt(neighbour, weight*rel->_weight2);
			base += weight*(1-rel->_weight2) + (1-weight)*rel->_weight2;
			xNeighbours.RemoveKey(neighbour);
		}else
			yNeighbours.SetAt(neighbour, rel->_weight2);
	}
	return (double)(xNeighbours.GetCount() + yNeighbours.GetCount())/(xNeighbours.GetCount() + yNeighbours.GetCount() + xyNeighbours.GetCount()*2);
}

double ORF::getNeighbourFCCount()
{
	CRBMap <FC*, int> map;
	POSITION pos;
	for(pos = _neighbours.GetHeadPosition(); pos;)
	{
		ORF *neighbour = _neighbours.GetNextKey(pos);
		FCList::iterator ft;
		for (ft = neighbour->_fc.begin(); ft != neighbour->_fc.end(); ft++)
			map.SetAt(*ft, 1);
	}
	return map.GetCount();
}

bool ORF::hasNeighbour(ORF* orf)
{
	return _neighbours.Lookup(orf)!=NULL;
}

//##################### FC MANAGER ########################

FCManager::FCManager()
{
}

FCManager::FCManager(const char *schemefile, const char *annotfile, const char *gridfile)
{
	ifstream sfile(schemefile);
	if (!sfile.is_open())
		throw 1;
	ifstream afile(annotfile);
	if (!afile.is_open())
		throw 2;
	ifstream gfile(gridfile);
	if (!gfile.is_open()) 
		throw 3;

	// read scheme file
	int index;
	char buffer[2048];
	string funcat, name;
	while (!sfile.eof())
	{
		if (!sfile.getline(buffer, sizeof(buffer)))
			break;
		if (buffer[0]=='#') continue;
		name = buffer;
		index = name.find(' ');
		funcat = name.substr(0, index);
		name = buffer+index+1;
		_InsertFC(funcat, name);
	}
	sfile.close();

	// read annot file
	FC *fc;
	int line = 0;
	while (!afile.eof())
	{
		line++;
		if (!afile.getline(buffer, sizeof(buffer)))
			continue;
		funcat = strupr(buffer);
		index = funcat.find('|');
		name = funcat.substr(0, index);
//		if (name.at(0)!='Y')
//			continue;

		funcat = funcat.substr(index+1, funcat.find('|', index+1)-index-1);
		if (!funcat.length()) continue;

		if (funcat.compare("799")==0 || 
			funcat.compare("96")==0 || 
			funcat.compare("98")==0 || 
			funcat.compare("99")==0)
		{
			ORF *newAnnot;
			if (!_activeORFs.Lookup(name, newAnnot))
			{
				newAnnot = new ORF(name);
				_activeORFs.SetAt(name, newAnnot);
			}
			continue;
		}

		fc = _FindFC(funcat);
		if (fc)
		{
			ORF* newAnnot;
			if (!_activeORFs.Lookup(name, newAnnot))
			{
				newAnnot = new ORF(name, fc);
				_activeORFs.SetAt(name, newAnnot);
				fc->_orfs.SetAt(newAnnot->_name, newAnnot);
			}else
			{
				if (newAnnot->hasFC(fc))
					cout << "Repeat Annotation: " << name << "|" <<  fc->getFC() << endl;
				else
				{
					newAnnot->_fc.push_back(fc);
					fc->_orfs.SetAt(newAnnot->_name, newAnnot);
				}
			}
		}else
		{
			cout << "line " << line << ":" << funcat << endl;
			_InsertFC(funcat, string("UNKNOWN"));
		}
	}
	afile.close();

	// read GRID file
	double weight;
	char token[255];
	buffer[0] = '\0';
	string orf1, orf2, type;
	ORF *a, *b;
//	while(!gfile.eof() && memcmp(buffer, "INTERACTOR_A", 12)!=0)
//		gfile.getline(buffer, sizeof(buffer));
	int intcount = 0;
	int repcount = 0;
	while(!gfile.eof())
	{
		gfile >> orf1 >> orf2;
		transform(orf1.begin(), orf1.end(), orf1.begin(), toupper);
//		if (orf1.at(0)!='Y')
//			continue;
		transform(orf2.begin(), orf2.end(), orf2.begin(), toupper);
//		if (orf2.at(0)!='Y')
//			continue;

		if (!gfile.getline(buffer, sizeof(buffer)))
			continue;
		
		// get experimental source
		index = 0;
		char *expt = trimleft(buffer);
//		tokenize(expt, '\t', index, token);
//		tokenize(expt, '\t', index, token);
		type = tokenize(expt, '\t', index, token);
		weight = 1;
		if (strlen(tokenize(expt, '\t', index, token))>0)
			weight = atof(token);
//		cout << weight << endl;

		if (!_activeORFs.Lookup(orf1, a))
		{
			a = new ORF(orf1);
			_activeORFs.SetAt(orf1, a);
		}
		if (!_activeORFs.Lookup(orf2, b))
		{
			b = new ORF(orf2);
			_activeORFs.SetAt(orf2, b);
		}
		if (a!=b)
		{
			repcount += a->addNeighbour(b, type, weight);
			b->addNeighbour(a, type, weight);
			intcount++;
		}
//		else
//			cout << "Cannot find " << orf1 << " - " << orf2 << endl;
	}
	cout << intcount << endl;
	cout << repcount << endl;
	cout << intcount - repcount << endl;
	gfile.close();
}

FC* FCManager::_FindFC(string &funcat)
{
	short id;
	int i;
	int depth = getcharcount('.', funcat)+1;
	FC *fc = NULL;
	FCMap *map = &_FCs;
	int start = 0;
	for (i=0; i<depth; i++)
	{
		int end = funcat.find('.', start);
		if (end<0) end = funcat.length()-1;
		id = atoi(funcat.substr(start, end-start+1).c_str());
		if (!map->Lookup(id, fc))
			return NULL;
		map = &fc->_children;
		start = end+1;
	}
	return fc;
}

void FCManager::_InsertFC(string &funcat, string &name)
{
	short id;
	int i;
	int depth = getcharcount('.', funcat)+1;
	FC *fc = NULL, *parent = NULL;
	FCMap *map = &_FCs;
	int start = 0;
	for (i=0; i<depth; i++)
	{
		int end = funcat.find('.', start);
		if (end<0) end = funcat.length()-1;
		id = atoi(funcat.substr(start, end-start+1).c_str());
		if (!map->Lookup(id, fc))
		{
			fc = new FC(id, name, parent);
			map->SetAt(id, fc);
			_informativeFCs.push_back(fc);
		}
//		else
//			cout << "error";
		map = &fc->_children;
		parent = fc;
		start = end+1;
	}
}

FCManager::~FCManager(void)
{
	POSITION pos = _activeORFs.GetHeadPosition();
	while(pos)
		delete (_activeORFs.GetNextValue(pos));

	pos = _FCs.GetHeadPosition();
	while(pos)
		delete (_FCs.GetNextValue(pos));
}

void FCManager::_getFCCount(int &count, FC *fc)
{
	short id;
	FC *childfc;
	FCMap *map;

	if (fc)
	{
		map = &fc->_children;
		if (fc->_orfs.GetCount() > 0)
			count++;
	}else
		map = &_FCs;

	POSITION pos = map->GetHeadPosition();
	while(pos)
	{
		map->GetNextAssoc(pos, id, childfc);
		_getFCCount(count, childfc);
	}
}

void FCManager::_PrintFCs(ofstream &ofile, string prefix, FC *fc)
{
	string newprefix;
	short id;
	FC *childfc;
	FCMap *map;

	if (fc)
	{
		map = &fc->_children;
		char id_string[10];
		sprintf(id_string, "%2d", fc->_id);
		newprefix = prefix + id_string;

		if (fc->_orfs.GetCount() > 0)
			ofile << newprefix << " (" << fc->_orfs.GetCount() << ")" << endl;
	}else
		map = &_FCs;

	POSITION pos = map->GetHeadPosition();
	while(pos)
	{
		map->GetNextAssoc(pos, id, childfc);
		_PrintFCs(ofile, newprefix + ":", childfc);
	}
}

int FCManager::_FilterUnInformative(int delta, FC *fc)
{
	int count = 0;
	bool child_ok = false;
	int childcount;
	short id;
	FC *childfc;
	FCMap *map;

	if (fc)
		map = &fc->_children;
	else
		map = &_FCs;

	POSITION pos = map->GetHeadPosition();
	while(pos)
	{
		map->GetNextAssoc(pos, id, childfc);
		childcount = _FilterUnInformative(delta, childfc);
		if (childcount >= delta)
		{
			child_ok = true;
			count += childcount;
		}
		else
		{
			POSITION npos = childfc->_orfs.GetHeadPosition();
			while(npos)
			{
				ORF* orf = childfc->_orfs.GetNextValue(npos);
				orf->removeFC(childfc);
				if (fc)
				{
					if (!orf->hasFC(fc))
						orf->_fc.push_back(fc);
					if (!fc->haveORF(orf))
						fc->_orfs.SetAt(orf->_name, orf);
				}
			}
			childfc->_orfs.RemoveAll();
		}
	}

	if (fc)
	{
		count += fc->_orfs.GetCount();

		if (child_ok)
		{
			POSITION npos = fc->_orfs.GetHeadPosition();
			while(npos)
			{
				ORF *orf = fc->_orfs.GetNextValue(npos);
				orf->removeFC(fc);
			}
			fc->_orfs.RemoveAll();
		}else
		if (count >= delta)
		{
			_informativeFCs.push_back(fc);
		}
	}
	return count;
}

void FCManager::RemoveUnInformative(int delta)
{
	_informativeFCs.clear();
	_FilterUnInformative(delta);
}

void FCManager::PrintAnnotations(const char *filename)
{
	ofstream ofile(filename);
	if (!ofile.is_open()) throw 1;

	POSITION pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF *annot = _activeORFs.GetNextValue(pos);
		FCList::iterator it = annot->_fc.begin();
		while(it != annot->_fc.end())
		{
			if ((*it)->_orfs.GetCount()>0)
			ofile << annot->_name << "|" << (*it)->getFC() << endl;
			it++;
		}
	}

	ofile.close();
}

void FCManager::PrintORFS(const char *filename)
{
	ofstream ofile(filename);
	if (!ofile.is_open()) throw 1;

	POSITION pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF *annot = _activeORFs.GetNextValue(pos);
		ofile << annot->_name << endl;
	}
	ofile.close();
}


void FCManager::PrintNeighbours(const char *filename)
{
	ofstream ofile(filename);
	if (!ofile.is_open()) throw 1;

	POSITION pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF *orf = _activeORFs.GetNextValue(pos);
		ofile << orf->_name << "\t" << orf->_neighbours.GetCount() <<"\t";
		POSITION pos1;
		for(pos1 = orf->_neighbours.GetHeadPosition(); pos1;)
		{
			NeighbourRelation *rel;
			ORF *neighbour;
			orf->_neighbours.GetNextAssoc(pos1, neighbour, rel);

			ofile << "|" << neighbour->_name << "-" << rel->_weight;
		}
		ofile << endl;
	}

	ofile.close();
}

void FCManager::Do_FunctionalFlow(ORF *annot)
{
	double potential1, potential2, score;
	string key;
	ORF *orf;
	FC *fc;
	FCList::iterator ft;
	POSITION npos, npos1;
	POSITION pos, pos1;
	CRBMap <FC*, double> functions;

	pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		_activeORFs.GetNextAssoc(pos, key, orf);
		orf->_probs.RemoveAll();
		orf->_probs_new.RemoveAll();
		for(ft = orf->_fc.begin(); ft != orf->_fc.end(); ft++)
			orf->_probs.SetAt(*ft, INFINITY);
	}

	for (int t=0; t<=6; t++)
	{
		pos = _activeORFs.GetHeadPosition();
		while(pos)
		{
			_activeORFs.GetNextAssoc(pos, key, orf);

			double base1 = 0;
			for (npos = orf->_neighbours.GetHeadPosition(); npos;)
				base1 += orf->_neighbours.GetNextValue(npos)->_weight2;
			for (npos = orf->_neighbours.GetHeadPosition(); npos;)
			{
				NeighbourRelation *rel;
				ORF *neighbour;
				orf->_neighbours.GetNextAssoc(npos, neighbour, rel);

				double base2 = 0;
				for (npos1 = neighbour->_neighbours.GetHeadPosition(); npos1;)
					base2 += neighbour->_neighbours.GetNextValue(npos1)->_weight2;
			
				for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
				{
					if (!orf->_probs.Lookup(*ft, potential1))
						potential1 = 0;
					if (!neighbour->_probs.Lookup(*ft, potential2))
						potential2 = 0;

					double inflow = 0;
					double outflow = 0;
					if (potential2 > potential1)
						inflow = min(rel->_weight2, rel->_weight2/base2);
					if (potential1 > potential2)
						outflow = min(rel->_weight2, rel->_weight2/base1);
					orf->_probs_new.SetAt(*ft, potential1 + inflow - outflow);
					
					if (!functions.Lookup(*ft, score))
						score = 0;
					score += inflow;
					functions.SetAt(*ft, score);
				}
			}
		}

		pos = _activeORFs.GetHeadPosition();
		while(pos)
		{
			_activeORFs.GetNextAssoc(pos, key, orf);

			orf->_probs.RemoveAll();
			pos1 = orf->_probs_new.GetHeadPosition();
			while(pos1)
			{
				orf->_probs_new.GetNextAssoc(pos1, fc, potential1);
				orf->_probs.SetAt(fc, potential1);
			}
			orf->_probs_new.RemoveAll();
		}
	}

	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
	{
		if (!functions.Lookup(*ft, potential1))
			potential1 = 0;
		(*ft)->_score = potential1;
	}
}

void FCManager::Do_NeighbourCounting(ORF *orf)
{
	FCList::iterator ft;
	POSITION pos, pos1;
	ScoreList temp;

	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
		(*ft)->_score = 0;

	for (pos = orf->_neighbours.GetHeadPosition(); pos;)
	{
		ORF *neighbour = orf->_neighbours.GetNextKey(pos);
		for (ft = neighbour->_fc.begin(); ft != neighbour->_fc.end(); ft++)
			(*ft)->_score += 1;
	}
	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
		if ((*ft)->_score > 0)
			temp.push_back(FCScore(*ft, (*ft)->_score));

	double score;
	CRBMap <FC*, double> map;
	temp.sort();
	temp.reverse();
	int i = 1;
	while (!temp.empty())
	{
		FCScore &fcscore = temp.front();
		map.SetAt(fcscore._fc, 1.0/i++);
		temp.pop_front();
	}
	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
	{
		if (map.Lookup(*ft, score))
			(*ft)->_score = score;
		else
			(*ft)->_score = 0;
	}

}

void FCManager::Do_Chi_Square(ORF *orf)
{
	FCList::iterator ft;
	POSITION pos, pos1;
	ScoreList temp;
	int base = 0;

	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
		(*ft)->_score = 0;

	for (pos = orf->_neighbours.GetHeadPosition(); pos;)
	{
		ORF *neighbour = orf->_neighbours.GetNextKey(pos);
		if (neighbour->_fc.size()>0)
			base++;
		for (ft = neighbour->_fc.begin(); ft != neighbour->_fc.end(); ft++)
			(*ft)->_score++;
	}

	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
	{
		double expected = (*ft)->_occ*base;
		if (expected > 0)
			(*ft)->_score  = pow((*ft)->_score - expected, 2)/expected;
		else
			(*ft)->_score  = 0;
		temp.push_back(FCScore(*ft, (*ft)->_score));
	}

	double score;
	CRBMap <FC*, double> map;
	temp.sort();
	temp.reverse();
	int i = 1;
	while (!temp.empty())
	{
		FCScore &fcscore = temp.front();
		map.SetAt(fcscore._fc, 1.0/i++);
		temp.pop_front();
	}
	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
	{
		if (map.Lookup(*ft, score))
			(*ft)->_score = score;
		else
			(*ft)->_score = 0;
	}
}

void FCManager::Do_Naive_Bayesian(ORF *orf)
{
	FCList::iterator ft;
	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
		(*ft)->_score = log(1-(*ft)->_occ);

	POSITION pos, pos1;
	for (pos = orf->_neighbours.GetHeadPosition(); pos;)
	{
		NeighbourRelation *rel;
		ORF *neighbour;
		orf->_neighbours.GetNextAssoc(pos, neighbour, rel);

		double prob = 0;
		double value;
		int count;
		string type;
		POSITION pos1 = rel->_types.GetHeadPosition();
		while(pos1)
		{
			rel->_types.GetNextAssoc(pos1, type, count);
			if (_typeprob.Lookup(type, value))
				prob += log(1-value);//*count;
		}
	
		for (ft = neighbour->_fc.begin(); ft != neighbour->_fc.end(); ft++)
			(*ft)->_score += log(1-rel->_weight);
		continue;

        for (pos1 = neighbour->_neighbours.GetHeadPosition(); pos1;)
		{
			NeighbourRelation *rel1;
			ORF *neighbour1;
			neighbour->_neighbours.GetNextAssoc(pos1, neighbour1, rel1);

			if (neighbour1 == orf)
				continue;
			for (ft = neighbour1->_fc.begin(); ft != neighbour1->_fc.end(); ft++)
				(*ft)->_score += log(1-rel->_weight*rel1->_weight);

		}
	}

	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
		(*ft)->_score = 1 - exp((*ft)->_score);
}

void FCManager::Do_FS_Weight(ORF *orf, bool background)
{
	FCList::iterator ft;
	double score;
	CRBMap <ORF*, double> neighbours;

	POSITION pos, pos1;
	for (pos = orf->_neighbours.GetHeadPosition(); pos;)
	{
		NeighbourRelation *rel;
		ORF *neighbour;
		orf->_neighbours.GetNextAssoc(pos, neighbour, rel);

		if (neighbour->_fc.size()>0)
		{
			if (!neighbours.Lookup(neighbour, score))
			{
				neighbour->_score = orf->computeFS_weight(neighbour, _activeORFs.GetCount(), true);
				score = 0;
			}
			score ++;
			neighbours.SetAt(neighbour, score);
			neighbour->_score = max(neighbour->_score, rel->_weight);
		}
//		continue;		
		for (pos1 = neighbour->_neighbours.GetHeadPosition(); pos1;)
		{
			NeighbourRelation *rel1;
			ORF *neighbour1;
			neighbour->_neighbours.GetNextAssoc(pos1, neighbour1, rel1);

			if (neighbour1 == orf)
				continue;
			if (neighbour1->_fc.size()>0)
			{
				if (!neighbours.Lookup(neighbour1, score))
				{
					neighbour1->_score = orf->computeFS_weight(neighbour1, _activeORFs.GetCount(), true);
					score = 0;
				}
				score ++;
				neighbours.SetAt(neighbour1, score);
				neighbour1->_score = max(neighbour1->_score, rel->_weight*rel1->_weight);
			}
		}

	}


	double base = 1;
	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
	{
		(*ft)->_score = background?(*ft)->_occ*g_reliability:0;
	}

	pos = neighbours.GetHeadPosition();
	while(pos)
	{
		ORF *neighbour;
		neighbours.GetNextAssoc(pos, neighbour, score);
		base += neighbour->_score*score;
		for (ft = neighbour->_fc.begin(); ft != neighbour->_fc.end(); ft++)
		{
			(*ft)->_score += neighbour->_score*score;
		}
	}
	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
	{
		(*ft)->_score /= base;
	}
}

void FCManager::RemoveFunctions(int num_fold, int fold)
{
	srand(100);//(unsigned)time(NULL));
	int i = 0;
	FCList::iterator ft;
	POSITION pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF *orf = _activeORFs.GetNextValue(pos);
		if (orf->_fc.size()==0)
			continue;
		orf->_backupfc = orf->_fc;
		double selection = ((double)rand()/RAND_MAX)*num_fold;
		if ((int)selection != fold)
			continue;
		orf->_fc.clear();
		i++;
	}
	cout << num_fold << "/" << fold << " " << i << " cleared" << endl;
//	getchar();
}

void FCManager::RestoreFunctions()
{
	FCList::iterator ft;
	POSITION pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF *orf = _activeORFs.GetNextValue(pos);
		if (orf->_backupfc.size()>0) 
		{
			orf->_fc = orf->_backupfc;
			orf->_backupfc.clear();
		}
	}
}


void FCManager::PrintUnknownPredicions(ofstream &ofile, short method, double significance)
{
	// predict
	int orfindex = 0;
	int orfcount = _activeORFs.GetCount();
	FCList::iterator it, ft;
	PredictScoreList::iterator st;
	POSITION pos;
	PredictScoreList predictions;

	prepareProbs();

	for (pos = _activeORFs.GetHeadPosition(); pos != NULL; orfindex++)
	{
		ORF *annot = _activeORFs.GetNextValue(pos);
			
		// Remove Annotation
		FCList temp;
		temp = annot->_fc;
		for (ft = annot->_fc.begin(); ft != annot->_fc.end(); ft++)
		{
			(*ft)->_orfs.RemoveKey(annot->_name);
			(*ft)->_occ = ((*ft)->_occ*_occ_base - 1)/(_occ_base - 1);
		}
		annot->_fc.clear();

		// Do Prediction
		cout << orfindex << " - " << annot->_name << endl;
		switch(method)
		{
		case NEIGHBOUR_COUNTING:
			Do_NeighbourCounting(annot);
			break;
		case CHI_SQUARE:
			Do_Chi_Square(annot);
			break;
		case FUNCTIONALFLOW:
			Do_FunctionalFlow(annot);
			break;
		case FS_WEIGHT:
			Do_FS_Weight(annot, true);
			break;
		case NAIVE_BAYESIAN:
			Do_Naive_Bayesian(annot);
			break;
		}

		// Restore Annotation
		annot->_fc = temp;
		for (ft = annot->_fc.begin(); ft != annot->_fc.end(); ft++)
		{
			(*ft)->_occ = ((*ft)->_occ*(_occ_base - 1) + 1)/_occ_base;
			(*ft)->_orfs.SetAt(annot->_name, annot);
		}
		for (it = _informativeFCs.begin(); it != _informativeFCs.end(); it++)
			if ((*it)->_score>0)
				predictions.push_back(PredictScore(annot, *it, (*it)->_score));
	}

	predictions.sort();
	predictions.reverse();
	
	int count = 0;
	int corrects = 0;
	for (st = predictions.begin(); st != predictions.end(); st++)
	{
		if (st->_orf->_fc.size()>0)
		{
			count++;
			if (st->_orf->hasFC(st->_fc))
				corrects++;
		}

		if (!st->_orf->hasFC(st->_fc))
		{
			double precision = 1;
			if (count>0)
				precision = (double)corrects/count;
			ofile << st->_score << "\t" << precision << "\t" << st->_orf->_name << "\t" << st->_fc->getFC() << "\t" << st->_fc->_name << endl;
		}
	}
}

void FCManager::PrintCrossValidate(ofstream &ofile, short method, bool LOO, double significance)
{
	// predict
	int orfindex = 0;
	int orfcount = _activeORFs.GetCount();
	FCList::iterator it, ft;
	ScoreList::iterator st;
	POSITION pos;

	for (pos = _activeORFs.GetHeadPosition(); pos != NULL;)
	{
		ORF *annot = _activeORFs.GetNextValue(pos);
		annot->_predictions.clear();
	}

	if (!LOO)
	{
		for (int i=0; i<NUM_FOLDS; i++)
		{
			cout << "FOLD " << i << endl;
			RemoveFunctions(NUM_FOLDS, i);
			prepareProbs();

			for (pos = _activeORFs.GetHeadPosition(), orfindex = 0; pos != NULL; orfindex++)
			{
				ORF *annot = _activeORFs.GetNextValue(pos);
				if (annot->_backupfc.size() == 0 || annot->_fc.size() > 0)
					continue;

				// Do Prediction
				cout << orfindex << " - " << annot->_name << endl;
				switch(method)
				{
				case NEIGHBOUR_COUNTING:
					Do_NeighbourCounting(annot);
					break;
				case CHI_SQUARE:
					Do_Chi_Square(annot);
					break;
				case FS_WEIGHT:
					Do_FS_Weight(annot, true);
					break;
				case FUNCTIONALFLOW:
					Do_FunctionalFlow(annot);
					break;
				case NAIVE_BAYESIAN:
					Do_Naive_Bayesian(annot);
					break;
				}

				for (it = _informativeFCs.begin(); it != _informativeFCs.end(); it++)
					if ((*it)->_score>0) annot->_predictions.push_back(FCScore(*it, (*it)->_score));
			}
			RestoreFunctions();
		}
	}else{
		prepareProbs();
		for (pos = _activeORFs.GetHeadPosition(); pos != NULL; orfindex++)
		{
			ORF *annot = _activeORFs.GetNextValue(pos);
			if (annot->_fc.size() == 0)
				continue;
			
			// Remove Annotation
			FCList temp;
			temp = annot->_fc;
			for (ft = annot->_fc.begin(); ft != annot->_fc.end(); ft++)
			{
				(*ft)->_orfs.RemoveKey(annot->_name);
				(*ft)->_occ = ((*ft)->_occ*_occ_base - 1)/(_occ_base - 1);
			}
			annot->_fc.clear();

			// Do Prediction
			cout << orfindex << " - " << annot->_name << endl;
			switch(method)
			{
			case NEIGHBOUR_COUNTING:
				Do_NeighbourCounting(annot);
				break;
			case CHI_SQUARE:
				Do_Chi_Square(annot);
				break;
			case FUNCTIONALFLOW:
				Do_FunctionalFlow(annot);
				break;
			case FS_WEIGHT:
				Do_FS_Weight(annot, true);
				break;
			case NAIVE_BAYESIAN:
				Do_Naive_Bayesian(annot);
				break;
			}

			// Restore Annotation
			annot->_fc = temp;
			for (ft = annot->_fc.begin(); ft != annot->_fc.end(); ft++)
			{
				(*ft)->_occ = ((*ft)->_occ*(_occ_base - 1) + 1)/_occ_base;
				(*ft)->_orfs.SetAt(annot->_name, annot);
			}
			for (it = _informativeFCs.begin(); it != _informativeFCs.end(); it++)
				if ((*it)->_score>0) annot->_predictions.push_back(FCScore(*it, (*it)->_score));
		}
	}

	int n = 0;
	ResultList scorelist;
	for (pos = _activeORFs.GetHeadPosition(); pos;)
	{
		ORF *annot = _activeORFs.GetNextValue(pos);
		n += annot->_fc.size();
		for (st = annot->_predictions.begin(); st != annot->_predictions.end(); st++)
			scorelist.push_back(Score(st->_score, annot->hasFC(st->_fc)));
	}
	// score
	scorelist.sort();
	scorelist.reverse();
	
	ResultList::iterator at;
	double threshold = 0.02;
	double lastvalue = scorelist.front()._score;
	int k = 0, m = 0;
	for (at=scorelist.begin(); at!=scorelist.end(); at++, m++)
	{
		if (at->_score != lastvalue)
		{
			if ((double)k/n > threshold)
			{
				ofile << lastvalue << "\t" << (double)k/n << "\t" << (double)k/m << endl;
				while(threshold < (double)k/n)
					threshold += 0.02;
			}
			lastvalue = at->_score;
		}
		if (at->_correct)
			k++;
	}
	if ((double)k/n > threshold)
		ofile << lastvalue << "\t" << (double)k/n << "\t" << (double)k/m << endl;
}

void FCManager::PrintFCs(const char *filename, int delta)
{
	ofstream ofile(filename);
	if (!ofile.is_open()) throw 1;
	string prefix;
	_PrintFCs(ofile, prefix);
	ofile.close();
}

int FCManager::_Flood(ORF *orf)
{
	int size = 1;
	int component = orf->_component;
	queue <ORF*> orfqueue;
	orfqueue.push(orf);
	
	POSITION pos;
	while(!orfqueue.empty())
	{
		ORF *nextorf = orfqueue.front();
		orfqueue.pop();

		for (pos = nextorf->_neighbours.GetHeadPosition(); pos;)
		{
			ORF *neighbour = nextorf->_neighbours.GetNextKey(pos);
			if (neighbour->_component == -1)
			{	
				neighbour->_component = component;
				orfqueue.push(neighbour);
				size++;
			}
		}
	}
	return size;
}

void FCManager::FilterIsolated()
{
	AnnotList deletions;
	AnnotList::iterator at;
	int changes = 1;
	while(changes>0)
	{
		changes = 0;
		POSITION pos = _activeORFs.GetHeadPosition();
		while(pos)
		{
			ORF *annot = _activeORFs.GetNextValue(pos);
			if (annot->getAnnotatedL1NeighbourCount()==0)
				deletions.push_back(annot);
		}
		changes = deletions.size();

		while(deletions.size()>0)
		{
			ORF *annot = deletions.front();
			deletions.pop_front();
			_activeORFs.RemoveKey(annot->_name);
			FCList::iterator ft;
			for (ft=annot->_fc.begin(); ft!=annot->_fc.end(); ft++)
				(*ft)->removeORF(annot);
			delete annot;
		}
	}
}

void FCManager::_FilterMinorComponents()
{
	POSITION pos = _activeORFs.GetHeadPosition();
	while(pos)
		(_activeORFs.GetNextValue(pos))->_component = -1;

	int size, maxsize = 0, connectedcomponents = 0, maxindex = 0;
	pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF * annot = _activeORFs.GetNextValue(pos);
		if (annot->_component == -1)
		{
			annot->_component = connectedcomponents;
			if ((size = _Flood(annot)) > maxsize)
			{
				maxsize = size;
				maxindex = connectedcomponents;
			}
//			cout << "Component " << connectedcomponents << ": " << size << endl;
			connectedcomponents++;
		}
	}
	cout << "Max Component : Component " << maxindex << " (" << maxsize << ")" << endl;

	queue <ORF*> deletions;
	pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF * annot = _activeORFs.GetNextValue(pos);
		if (annot->_component != maxindex)
			deletions.push(annot);
	}
	while(!deletions.empty())
	{
		ORF *annot = deletions.front();
		_activeORFs.RemoveKey(annot->_name);
		delete annot;
		deletions.pop();
	}
}

void FCManager::PrintProdistinMatrix()
{
	_FilterMinorComponents();

	FILE *matrixfile = fopen("distmatrix.txt", "wt");

	fprintf(matrixfile, "%d\n",  _activeORFs.GetCount());
	
	POSITION pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF *x = _activeORFs.GetNextValue(pos);
		fprintf(matrixfile, "%s\n", x->_name.c_str());

		POSITION pos1 = _activeORFs.GetHeadPosition();
		if (pos1)
			fprintf(matrixfile, "%lf", x->computeCD_dist_OLD(_activeORFs.GetNextValue(pos1)));
		while(pos1)
			fprintf(matrixfile, " %lf", x->computeCD_dist_OLD(_activeORFs.GetNextValue(pos1)));
		fprintf(matrixfile, "\n"); 
	}
	fclose(matrixfile);
}

void FCManager::PrintProdistin()
{
	_FilterMinorComponents();

	AnnotTree tree;
	tree.readTree("loc_tree.txt", _activeORFs);
//	tree.writeTree("test.txt");

	ofstream predict("loc_out.txt");
	ofstream ofile("results.txt");

	int correct[20][9], wrong[20][9], valid[20], solution[9];
	for (int i=0; i<20; i++)
	{
		valid[i] = 0;
		for (int j=0; j<9; j++)
		{
			correct[i][j] = 0;
			wrong[i][j] = 0;
			solution[j] = 0;
		}
	}

	int n = 0;
	ResultList resultlist;
	ScoreList scorelist;
	ScoreList::iterator ft;
	POSITION pos;
	for(pos = _activeORFs.GetHeadPosition(); pos!=NULL;)
	{
		ORF *orf = _activeORFs.GetNextValue(pos);
		cout << orf->_name << endl;
		if (orf->_fc.size()==0)
			continue;
		n += orf->_fc.size();
		predict << orf->_name << endl;
		scorelist = tree.predict(orf);
		scorelist.sort();
		scorelist.reverse();
		for (ft = scorelist.begin(); ft != scorelist.end(); ft++)
		{
			predict << ft->_fc->getFC() << "\t" << ft->_score << endl;
			resultlist.push_back(Score(ft->_score, orf->hasFC(ft->_fc)));
		}

		int i=0;
		ScoreList::iterator ft;
		int degree = orf->_neighbours.GetCount();
		if (degree>7)
			degree = 8;
		if (scorelist.size()>0)
			solution[degree]++;
		for (ft=scorelist.begin(); ft!=scorelist.end(); ft++)
		{
//			ofile << ft->_fc->getFC() << "\t" << ft->_score << endl;
			if (orf->hasFC(ft->_fc))
			{
				for (int j=i; j<20; j++)
					correct[j][degree]++;
				i = 20;
				break;
			}else
				wrong[i][degree]++;
			if(++i == 20)
				break;
		}
		if (i < 20)
		{
			for (int j=i; j<20; j++)
				wrong[j][degree]++;			
		}

		i = 0;
		for (ft=scorelist.begin(); ft!=scorelist.end(); ft++)
		{
			if (orf->hasFC(ft->_fc))
			{
				for (int j=i; j<20; j++)
					valid[j]++;
			}
			if(++i == 20)
				break;
		}
	}
	predict.close();

	ofile << "Degree\tTotal";
	int corrects[20], total = 0, count, solutions = 0;
	for (int i=0; i<20; i++)
	{
		corrects[i] = 0;
		ofile << "\tTop " << i+1;
	}
	ofile << endl;

	for (int i=1; i<8; i++)
	{
		count = correct[0][i] + wrong[0][i];
		total += count;
		ofile << i << "\t" << count;
		for (int j=0; j<20; j++)
		{
			corrects[j] += correct[j][i];
			ofile << "\t" << (double)correct[j][i]/count;
		}
		solutions += solution[i];
		ofile << "\t" << solution[i];
		ofile << endl;
	}
	count = correct[0][8]+wrong[0][8];
	total += count;
	ofile << ">7\t" << count;
	for (int j=0; j<20; j++)
	{
		corrects[j] += correct[j][8];
		ofile << "\t" << (double)correct[j][8]/count;
	}
	solutions += solution[8];
	ofile << "\t" << solution[8];
	ofile << endl;

	ofile << "All\t" << total;
	for (int j=0; j<20; j++)
		ofile << "\t" << (double)corrects[j]/total;
	ofile << endl;
	ofile << "Corrects\t" << total;
	for (int j=0; j<20; j++)
		ofile << "\t" << valid[j];
	ofile << "\t" << solutions;
	ofile << endl;

	// score
	resultlist.sort();
	resultlist.reverse();
		
	ResultList::iterator at;
	double threshold = 0.02;
	double lastvalue = scorelist.front()._score;
	int k = 0, m=0;
	for (at=resultlist.begin(); at!=resultlist.end(); at++, m++)
	{
		if (at->_score != lastvalue)
		{
			if ((double)k/n > threshold)
			{
				ofile << at->_score << "\t" << (double)k/n << "\t" << (double)k/m << endl;
				while(threshold < (double)k/n)
					threshold += 0.02;
			}
			lastvalue = at->_score;
		}
		if (at->_correct)
			k++;
	}

	ofile.close();
}

TreeNode* AnnotTree::_readSubTree(ifstream &treefile, AnnotMap &map, TreeNode *parent)
{
	TreeNode* newnode = new TreeNode(parent, NULL);
	char ch, buffer[255];
	treefile >> ch;
	while (!treefile.eof())
	{
		switch(ch)
		{
		case '(':
			newnode->_children.push_back(_readSubTree(treefile, map, newnode));
			treefile >> ch;
			break;

		case ':':
			while (ch != ',' && ch != ')')
				treefile >> ch;
			if (ch == ',')
			{
				treefile >> ch;
				break;
			}

		case ')':
		case ';':
		case '\n':
			return newnode;

		default:
			int index = 0;
			buffer[index++] = ch;
			while(ch != ':')
			{
				treefile >> ch;
				buffer[index++] = ch;
			}
			buffer[index-1] = '\0';
			ORF *annot;
			cout << buffer << endl;
			if (map.Lookup(buffer, annot))
				newnode->_children.push_back(new TreeNode(newnode, annot));
			else
				cout << "error - " << buffer << endl;
			break;
		}
	}
	return newnode;
}

void AnnotTree::_writeSubTree(ofstream &treefile, TreeNode *subtree)
{
	if (subtree->_annot)
		treefile << subtree->_annot->_name;
	else
	{
		treefile << "(";
		TreeNodeList::iterator it = subtree->_children.begin();
		if (it != subtree->_children.end())
		{
			_writeSubTree(treefile, *it);
			it++;			
		}
		while(it != subtree->_children.end())
		{
			treefile << ",";
			_writeSubTree(treefile, *it);
			it++;
		}
		treefile << ")";
	}
}

void AnnotTree::readTree(char *filename, AnnotMap &map)
{
	ifstream treefile(filename);
	char ch;
	treefile >> ch;
	_root = _readSubTree(treefile, map, NULL);
	treefile.close();
}

void AnnotTree::writeTree(char *filename)
{
	ofstream treefile(filename);
	_writeSubTree(treefile, _root);
	treefile.close();
}

TreeNode* AnnotTree::_findNode(ORF *annot, TreeNode* subtree)
{
	if (subtree->_annot == annot)
		return subtree;
	else
	{
		TreeNode *temp;
		TreeNodeList::iterator it = subtree->_children.begin();
		while(it != subtree->_children.end())
		{
			if ((temp = _findNode(annot, *it)))
				return temp;
			it++;
		}
		return NULL;
	}
}

void AnnotTree::_getAnnots(TreeNode* subtree, AnnotList &list)
{
	if (subtree->_annot && subtree->_annot->_fc.size() > 0)
		list.push_back(subtree->_annot);
	else
	{
		TreeNodeList::iterator it = subtree->_children.begin();
		while(it != subtree->_children.end())
		{
			_getAnnots(*it, list);
			it++;
		}
	}
}

ScoreList AnnotTree::predict(ORF *annot)
{
	TreeNode* node = _findNode(annot, _root);
	if (!node)
		cout << "aaa" << endl;

	int count;
	ScoreList list;
	FC *selected = NULL, *fc;
	node = node->_parent;
	while(node)
	{
		AnnotList children;
		AnnotList::iterator at;
		_getAnnots(node, children);
		CRBMap <FC*, int> map;
		
		int childrensize = 0;
		for (at = children.begin(); at != children.end(); at++)
		{
			ORF *child = *at;
			if (child == annot)
				continue;
			FCList::iterator ft;
			for (ft = child->_fc.begin(); ft != child->_fc.end(); ft++)
			{
				if (!map.Lookup(*ft, count))
					count = 0;
				count++;
				map.SetAt(*ft, count);
			}
			childrensize++;
		}

		ScoreList temp;
		POSITION pos;
		for (pos = map.GetHeadPosition(); pos != NULL;)
		{
			map.GetNextAssoc(pos, fc, count);
			if (count >=3 && ((double)count/childrensize)>=0.5)
				temp.push_back(FCScore(fc, (double)count/childrensize));
		}
		if (temp.size()>0)
			list = temp;
		node = node->_parent;
	}
	return list;
}

int FCManager::GetInteractionCount()
{
	short count;
	int interactions = 0;
	CRBMap <ORF*, short> map;
	POSITION pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF *i = _activeORFs.GetNextValue(pos);
		POSITION npos;
		for (npos = i->_neighbours.GetHeadPosition(); npos;)
		{
			ORF *j = i->_neighbours.GetNextKey(npos);
			if (!map.Lookup(j, count))
				interactions++;
		}
		map.SetAt(i, 1);
	}
	return interactions;
}

void FCManager::prepareProbs()
{
	// compute evidence weight
	int index;
//		cout << "Computing reliability weights..." << endl;
	string key;
	double weight;
	POSITION npos, npos1;
	CRBMap <ORF *, int> map;
	// clear values
	_typeprob.RemoveAll();
	_typeocc.RemoveAll();

	POSITION pos;
	pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF *orf = _activeORFs.GetNextValue(pos);
		if (orf->_fc.size()>0)
		{
			for (npos = orf->_neighbours.GetHeadPosition(); npos;)
			{
				NeighbourRelation *rel;
				ORF *neighbour;
				orf->_neighbours.GetNextAssoc(npos, neighbour, rel);
				
				if (map.Lookup(neighbour, index) || neighbour->_fc.size()==0)
					continue;
				POSITION pos1 = rel->_types.GetHeadPosition();
				while(pos1)
				{
					rel->_types.GetNextAssoc(pos1, key, index);
					
					if (!_typeprob.Lookup(key, weight))
						weight = 0;
					if (orf->getSharedFCcount(neighbour)>0)
						weight++;
					_typeprob.SetAt(key, weight);
					if (!_typeocc.Lookup(key, index))
						index = 0;
					index ++;
					_typeocc.SetAt(key, index);
				}
			}
		}
		map.SetAt(orf, 1);
	}
	pos = _typeprob.GetHeadPosition();
	while(pos)
	{
		_typeprob.GetNextAssoc(pos, key, weight);
		if (!_typeocc.Lookup(key, index))
		continue;
		double rel = weight/index;
//			if (rel <0.18) rel = 0;
		_typeprob.SetAt(key, rel);
		cout << key << "\t" << weight << " / " << index << " = " << weight/index << "(" << rel << ")" << endl;
	}
//		cout << "OK" << endl;

	// estimate lambda
	int size;
	int known = 0;
	int good = 0;
	int total = 0;
	int links = 0;
	pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF *orf = _activeORFs.GetNextValue(pos);
		if (orf->_neighbours.GetCount()==0)
			continue;
		total++;
		links += orf->_neighbours.GetCount();
		if (orf->_fc.size()==0 || !(size = orf->getAnnotatedL1NeighbourCount()))
			continue;
		for (npos = orf->_neighbours.GetHeadPosition(); npos;)
		{
			ORF *neighbour = orf->_neighbours.GetNextKey(npos);
			if (neighbour->_fc.size()==0)
				continue;
			known ++;
			if (orf->getSharedFCcount(neighbour)>0)
				good ++;
		}
	}
	_interactions = links/2;
	double reliability = (double)good/known;
	g_lambda = (double)links/total;
	g_lambda_r = reliability*g_lambda;
	g_reliability = reliability;
	cout << g_lambda << endl;

	// compute FS Sim R
	pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF *orf = _activeORFs.GetNextValue(pos);
		for (npos = orf->_neighbours.GetHeadPosition(); npos;)
		{
			NeighbourRelation *rel;
			ORF *neighbour;
			orf->_neighbours.GetNextAssoc(npos, neighbour, rel);
			int count;
			string type;
			double weight = 1, value;
			POSITION pos1 = rel->_types.GetHeadPosition();
			while(pos1)
			{
				rel->_types.GetNextAssoc(pos1, type, count);
				if (_typeprob.Lookup(type, value))
					weight *= pow(1-value, count);
			}
			rel->_weight2 = 1 - weight;
		}
	}

//	ofstream PRfile("PR.txt");
	map.RemoveAll();
	pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF *orf = _activeORFs.GetNextValue(pos);
//		cout << orf->_name << endl;
		for (npos = orf->_neighbours.GetHeadPosition(); npos;)
		{
			NeighbourRelation *rel, *rel1;
			ORF *neighbour;
			orf->_neighbours.GetNextAssoc(npos, neighbour, rel);
			if (map.Lookup(neighbour))
				continue;
			neighbour->_neighbours.Lookup(orf, rel1);
			rel->_weight = rel1->_weight = orf->computeFS_weight(neighbour, _activeORFs.GetCount(), true);
		}
		map.SetAt(orf, 1);
	}
//	PRfile.close();

	_occ_base = 0;
	_occ_int_base = 0;
	FCList::iterator ft;
	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
	{
		(*ft)->_occ = 0;
		(*ft)->_occ_int = 0;
	}
	pos = _activeORFs.GetHeadPosition();
	while(pos)
	{
		ORF *orf = _activeORFs.GetNextValue(pos);
		if (orf->_fc.size() > 0)
		{
			_occ_base ++;
		
			for (ft = orf->_fc.begin(); ft != orf->_fc.end(); ft++)
				(*ft)->_occ++;
		}
		POSITION pos1;
		for(pos1 = orf->_neighbours.GetHeadPosition(); pos1;)
		{
            ORF *neighbour = orf->_neighbours.GetNextKey(pos1);
			if (orf->_fc.size() > 0 || neighbour->_fc.size() > 0 )
				_occ_int_base ++;
			else
				continue;

			for (ft = orf->_fc.begin(); ft != orf->_fc.end(); ft++)
				(*ft)->_occ_int++;
			for (ft = neighbour->_fc.begin(); ft != neighbour->_fc.end(); ft++)
				(*ft)->_occ_int++;
		}
	}
		
	for (ft = _informativeFCs.begin(); ft != _informativeFCs.end(); ft++)
	{
		(*ft)->_occ /= _occ_base;
		(*ft)->_occ_int /= _occ_int_base;
	}
//		_avg /= _informativeFCs.size()*2;
}

int FCManager::GetAnnotCount()
{
	int count = 0;
	POSITION pos = _activeORFs.GetHeadPosition();
	while(pos)
		count += (_activeORFs.GetNextValue(pos))->_fc.size();
	return count;
}

int FCManager::GetAnnotated()
{
	int count = 0;
	POSITION pos = _activeORFs.GetHeadPosition();
	while(pos)
		count += ((_activeORFs.GetNextValue(pos))->_fc.size()>0?1:0);
	return count;
}
