#include <windows.h>
#include <iostream>
#include <fstream>
#include "FCManager.h"

void main(int argc, const char** argv)
{	
	if (argc < 3)
	{
		int suffixlen = strlen(strrchr(argv[0], '.'));
		char *commandname = strrchr(argv[0], '\\') + 1;
		commandname[strlen(commandname)-suffixlen] = '\0';
		cout << "usage: " << commandname << " schemefile annotationfile interactionfile" << endl;
		getchar();
		return;
	}
	cout << "Scheme: " << argv[1] << endl;
	cout << "Annotations: " << argv[2] << endl;
	cout << "Interactions: " << argv[3] << endl << endl;

	try{
		FCManager fc_man(argv[1], argv[2], argv[3]);
		cout << "Total FCs:" << fc_man.GetFCCount() << endl;
		cout << "Total ORFs:" << fc_man.GetOrfCount() << endl;
		cout << "Total Annotations:" << fc_man.GetAnnotCount() << endl;
		cout << "Total Annoted ORFs:" << fc_man.GetAnnotated() << endl;
		cout << "Total Interactions:" << fc_man.GetInteractionCount() << endl;		

		// Remove non informative FCs
		fc_man.RemoveUnInformative(30);		// Remove uninformatuve Functional Classes
		fc_man.FilterIsolated();			// Remove proteins with no known neighbours

		cout << "Total Proteins:" << fc_man.GetOrfCount() << endl;
		cout << "Functional Classes:" << fc_man.GetFCCount() << endl;
		cout << "Annotated Proteins:" << fc_man.GetAnnotated() << endl;
		cout << "Total Annotations:" << fc_man.GetAnnotated() << endl;

		ofstream ofile("results.txt");

		// Do cross validation
		fc_man.PrintCrossValidate(ofile, FS_WEIGHT, true);				// FS-Weighted Avg Leave One Out CV
//		fc_man.PrintCrossValidate(ofile, FS_WEIGHT, false);				// FS-Weighted Avg NUMFOLD Cross Validation
//		fc_man.PrintCrossValidate(ofile, NEIGHBOUR_COUNTING, false);	// Neighbour Counting
//		fc_man.PrintCrossValidate(ofile, CHI_SQUARE, false);			// Chi Square
//		fc_man.PrintCrossValidate(ofile, FUNCTIONALFLOW, false);		// Functional Flow
		ofile.close();

		// Do PRODISTIN Clustering
//		fc_man.PrintProdistinMatrix();									
		// Output distance matrix to distmatrix.txt for clustering with BIONJ
//		fc_man.PrintProdistin();
		// Read clustered tree tree.txt from BIONJ and print predictions to out.txt, precision vs recall to results.txt

	}catch(int a)
	{
		switch (a)
		{
		case 1:
			cout << "Function Scheme file unavailable!" << endl;
			break;
		case 2:
			cout << "Function Annotation file unavailable!" << endl;
			break;
		case 3:
			cout << "Interaction file unavailable!" << endl;
			break;	
		};
		getchar();
	}
//	getchar();
}